{------------------------------------------------------------------------------
  plsLangMan.pas

  Precision Language Suite for VCL

  written by  Precision software & consulting
              e-mail: info@be-precision.com
              web: http://www.be-precision.com

  Purpose:    Localization support
              (main part of Precision Language Suite for VCL)

  The source code is given as is. The author is not responsible
  for any possible damage done due to the use of this code.
  This unit can be freely used in any application. The complete
  source code remains property of the author and may not be distributed,
  published, given or sold in any form as such. No parts of the source
  code can be included in any other component or application without
  written authorization of the author.

  Copyright  2008 - 2011  Precision software & consulting
  All rights reserved

------------------------------------------------------------------------------
  History:

  - Version: 2.2.4.19
    * added - BiDiMode property
    * added - ActiveFiles property

  - Version: 2.1.3.17
    * added - Support for localizing the resourcestrings (see LangResourceStr method and LangSuiteDemo\LangConsts.pas).
              It is recommended to use the LangConsts.pas code (or similar) for localizing the resourcestrings,
              because it ensures, that the localized text is stored (duplicated) in a persistant variable, that avoids a memory leaks.
    * added - Support for reading language files in LNGZ format (pk-zip compatible).
              You have to define PLS_LNGZ conditional directive to use it.

  - Version: 2.0.2.9
    * fixed - GetAvailableLanguages for Delphi 2007 and below (reading UTF-8 encoded files)
------------------------------------------------------------------------------}

unit plsLangMan;

{$I plsCompilers.Inc}
{$M+}

interface

uses Windows, Messages, Classes, IniFiles;

const
  // Internal message identifier to process the language change in child forms/frames/modules
  WM_LANGUAGECHANGED = WM_APP + 512;
  LANGFILEEXT = 'lng';
  LANGFILEEXTZ = 'lngz';
  LANGFILEEXTU = 'lngu';
  // Language file main section
  lini_sec_main = 'precision language file';
  // Language file constants section
  lini_sec_consts = 'constants';
  // Language file special constants that can be converted from source strings given at runtime via masks
  lini_sec_messages = 'messages';
  // Key for reading the language name (title)
  lini_key_name = 'languagename';
  // Key for "right to left" (bidirectional mode) flag
  lini_key_rtl = 'rtl';

type
  // Main class of Precision Language Suite
	TLanguageManager=class
	private
    fAllowSpecialProps:Boolean;
    fOnLanguageChanged:TNotifyEvent;
    fRecursiveFolder:Boolean;
    fActiveFiles:TStringList;
    fFileNames:TStringList;
    fFolder:string;
    fLanguageCode:string;
    fLanguageName:string;
    fBiDiMode:TBiDiMode;
		fLangProps:TStringList;
		fLangConsts:TStringList;
    fEMessList:TStringList;
    fEMessConv:TStringList;
    fLangResStr:TStringList;
    fLastError:string;
    procedure SetLanguageCode(const Value:string);
    procedure FilterFileNames(const FileList:TStrings);
    function ReadLanguageData(Ini:TMemIniFile; IsUser:Boolean; var LangName:string):Boolean;
    function LoadLanguageFile(const FileName:string;var LangName:string;var ErrMsg:string):Boolean;
	public
    constructor Create;
		destructor Destroy; override;
    // Clears props, consts and eventually the messages. Resourcestring constants (started with @) are always cleared.
    procedure Clear(Consts:Boolean=true;Props:Boolean=True;Messages:Boolean=False);
    // Returns all available languages in form (langcode=langname)
    function GetAvailableLanguages(const AvailableLanguages:TStringList):Boolean;
    procedure GetLanguageFiles(LngCode:string; const PdkFiles, UsrFiles:TStrings);
    // Reloads current language
    procedure Refresh;
    // Returns localized constant or property
    function LangText(const ID:string;const DefaultText:string=''):string;
    // Localize the whole form/frame/module
    function LangVCL(const Owner:TComponent;const AsClassName:String='';
        const OnlyComponent:TComponent=nil):Boolean;
    // Returns localized message converted via mask or returns itself
    function LangMess(const aMessage:string):string;
    // Localize the specified resourcestring (use only inside the LanguageChanged event)
    procedure LangResourceStr(rs: PResStringRec; newStr: PChar);
  published
    // List of currently loaded files (including the full path). The list is filled automatically by setting the LanguageCode property.
    property ActiveFiles:TStringList read fActiveFiles;
    // Files (filenames without an extensions) to load. By this list you can filter the language files to be loaded.
    property FileNames:TStringList read fFileNames;
    // Folder with language files ('AppExe\Langs' as default)
    property Folder:string read fFolder write fFolder;
    // Search recursively in language folder
    property RecursiveFolder:Boolean read fRecursiveFolder write fRecursiveFolder default False;
    // User friendly name of language
    property LanguageName:string read fLanguageName;
    // Loads given language [Primary or secondary language code (i.e. 'en','cs','US','CZ','GB',...) = first language file extension]
    property LanguageCode:string read fLanguageCode write SetLanguageCode;
    // BiDi mode for current language
    property BiDiMode:TBiDiMode read fBiDiMode;
    // List of all properties
    property Properties:TStringList read fLangProps;
    // List of all constants
    property Constants:TStringList read fLangConsts;
    // If true then allow usage of properties like TShortCut, but applying language is a little slower
    property AllowSpecialProps:Boolean read fAllowSpecialProps write fAllowSpecialProps;
    // Holds the last LanguageManager processing error
    property LastError:string read fLastError;
    // Triggered after loading the language
    property OnLanguageChanged:TNotifyEvent read fOnLanguageChanged write fOnLanguageChanged;
	end;

var
  // Predefined global instance of TLanguageManager
	LanguageManager:TLanguageManager;

// Returns default language code for current locale
function GetDefaultLangCode:string;
// Returns primary language code for given locale
function GetLangPrimaryCode(aLangID:LANGID):String;
// Returns the language subcode for given locale
function GetLangSubCode(aLangID:LANGID):string;
// Returns the whole language code (ie. "en-US") for given locale
function GetLangCountry(aLangID:LANGID):string;

// Returns the text encoded by escape sequences for CRLF, TAB and other special characters
function _ECodeText(const Txt:string):string;

implementation

uses
  SysUtils, TypInfo, Masks, {$IFDEF D6}StrUtils,{$ENDIF D6} {$IFDEF PLS_LNGZ}SciZipFile,{$ENDIF} Menus;

const
  LANGDIR_RELPATH = 'Langs';

  CRLF = #13#10;
  Tab  = #9;
  V_CRLF = '\n';
  V_TAB = '\t';

  // Self object reference identifier to localize form/frame/module properties
  LC_Self = 'self';

  resELangFilesNotFound = 'Language not found';

{-----------------------------------------------------------------------------
  SUPPORT ROUTINES
-----------------------------------------------------------------------------}

function GetDefaultLangCode:String;
begin
  Result:=GetLangPrimaryCode(GetUserDefaultLangID)
end;

function GetLangPrimaryCode(aLangID:LANGID):String;
begin
  case aLangID of
    $0436	: Result := 'af';
    $041c	: Result := 'sq';
    $0484	: Result := 'gsw';
    $045e	: Result := 'am';
    $1401,$3c01,$0c01,$0801,$2c01,$3401,$3001,$1001,$1801,$2001,$4001,$0401,$2801,$1c01,$3801,$2401
          : Result := 'ar';
    $042b	: Result := 'hy';
    $044d	: Result := 'as';
    $082c,$042c
        	: Result := 'az';
    $046d	: Result := 'ba';
    $042d	: Result := 'eu';
    $0423	: Result := 'be';
    $0445	: Result := 'bn';
    $781a, $201a, $141a
        	: Result := 'bs';
    $047e	: Result := 'br';
    $0402	: Result := 'bg';
    $0403	: Result := 'ca';
    $0c04,$1404,$1004,$0804,$0404
          : Result := 'zh';
    $0483	: Result := 'co';
    $041a,$101a
        	: Result := 'hr';
    $0405	: Result := 'cs';
    $0406	: Result := 'da';
    $048c	: Result := 'gbz';
    $0465	: Result := 'dv';
    $0813,$0413
        	: Result := 'nl';
    $0c09,$2809,$1009,$2409,$4009,$1809,$2009,$4409,$1409,$3409,$4809,$1c09,$2c09,$0809,$0409,$3009
          : Result := 'en';
    $0425	: Result := 'et';
    $0438	: Result := 'fo';
    $0464	: Result := 'fil';
    $040b	: Result := 'fi';
    $080c,$0c0c,$040c,$140c,$180c,$100c
        	: Result := 'fr';
    $0462	: Result := 'fy';
    $0456	: Result := 'gl';
    $0437	: Result := 'ka';
    $0c07,$0407,$1407,$1007,$0807
        	: Result := 'de';
    $0408	: Result := 'el';
    $046f	: Result := 'kl';
    $0447	: Result := 'gu';
    $0468	: Result := 'ha';
    $040d	: Result := 'he';
    $0439	: Result := 'hi';
    $040e	: Result := 'hu';
    $040f	: Result := 'is';
    $0470	: Result := 'ig';
    $0421	: Result := 'id';
    $085d,$045d
        	: Result := 'iu';
    $083c	: Result := 'ga';
    $0410,$0810
        	: Result := 'it';
    $0411	: Result := 'ja';
    $044b	: Result := 'kn';
    $043f	: Result := 'kk';
    $0453	: Result := 'kh';
    $0486	: Result := 'qut';
    $0487	: Result := 'rw';
    $0457	: Result := 'kok';
    $0412	: Result := 'ko';
    $0440	: Result := 'ky';
    $0454	: Result := 'lo';
    $0426	: Result := 'lv';
    $0427	: Result := 'lt';
    $082e	: Result := 'dsb';
    $046e	: Result := 'lb';
    $042f	: Result := 'mk';
    $083e,$043e
        	: Result := 'ms';
    $044c	: Result := 'ml';
    $043a	: Result := 'mt';
    $0481	: Result := 'mi';
    $047a	: Result := 'arn';
    $044e	: Result := 'mr';
    $047c	: Result := 'moh';
    $0450,$0850
        	: Result := 'mn';
    $0461	: Result := 'ne';
    $0414,$0814
        	: Result := 'no';
    $0482	: Result := 'oc';
    $0448	: Result := 'or';
    $0463	: Result := 'ps';
    $0429	: Result := 'fa';
    $0415	: Result := 'pl';
    $0416,$0816
        	: Result := 'pt';
    $0446	: Result := 'pa';
    $046b,$086b,$0c6b
        	: Result := 'quz';
    $0418	: Result := 'ro';
    $0417	: Result := 'rm';
    $0419	: Result := 'ru';
    $243b,$103b,$143b,$0c3b,$043b,$083b,$203b,$183b,$1c3b
        	: Result := 'se';
    $044f	: Result := 'sa';
    $7c1a,$181a,$0c1a,$081a
        	: Result := 'sr';
    $046c	: Result := 'ns';
    $0432	: Result := 'tn';
    $045b	: Result := 'si';
    $041b	: Result := 'sk';
    $0424	: Result := 'sl';
    $2c0a,$400a,$340a,$240a,$140a,$1c0a,$300a,$440a,$100a,$480a,$080a,$4c0a,$180a,$3c0a,$280a,$500a,$0c0a,$380a,$200a
        	: Result := 'es';
    $0441	: Result := 'sw';
    $081d,$041d
        	: Result := 'sv';
    $045a	: Result := 'syr';
    $0428	: Result := 'tg';
    $085f	: Result := 'tmz';
    $0449	: Result := 'ta';
    $0444	: Result := 'tt';
    $044a	: Result := 'te';
    $041e	: Result := 'th';
    $0451	: Result := 'bo';
    $041f	: Result := 'tr';
    $0442	: Result := 'tk';
    $0480	: Result := 'ug';
    $0422	: Result := 'uk';
    $042e	: Result := 'wen';
    $0820,$0420
        	: Result := 'ur';
    $0843,$0443
        	: Result := 'uz';
    $042a	: Result := 'vi';
    $0452	: Result := 'cy';
    $0488	: Result := 'wo';
    $0434	: Result := 'xh';
    $0485	: Result := 'sah';
    $0478	: Result := 'ii';
    $046a	: Result := 'yo';
    $0435	: Result := 'zu';
  else
    Result:='';
  end;
end;

function GetLangSubCode(aLangID:LANGID):String;
begin
  case aLangID of
    $0436	: Result := 'za';
    $041c	: Result := 'al';
    $0484	: Result := 'fr';
    $045e	: Result := 'et';
    $1401	: Result := 'dz';
    $3c01	: Result := 'bh';
    $0c01	: Result := 'eg';
    $0801	: Result := 'iq';
    $2c01	: Result := 'jo';
    $3401	: Result := 'kw';
    $3001	: Result := 'lb';
    $1001	: Result := 'ly';
    $1801	: Result := 'ma';
    $2001	: Result := 'om';
    $4001	: Result := 'qa';
    $0401	: Result := 'sa';
    $2801	: Result := 'sy';
    $1c01	: Result := 'tn';
    $3801	: Result := 'ae';
    $2401	: Result := 'ye';
    $042b	: Result := 'am';
    $044d	: Result := 'in';
    $082c,$042c
          : Result := 'az';
    $046d	: Result := 'ru';
    $042d	: Result := 'es';
    $0423	: Result := 'by';
    $201a,$141a
          : Result := 'ba';
    $047e	: Result := 'fr';
    $0402	: Result := 'bg';
    $0403	: Result := 'es';
    $0c04	: Result := 'hk';
    $1404	: Result := 'mo';
    $1004	: Result := 'sg';
    $0483	: Result := 'fr';
    $101a	: Result := 'ba';
    $041a	: Result := 'hr';
    $0405	: Result := 'cz';
    $0406	: Result := 'dk';
    $048c	: Result := 'af';
    $0465	: Result := 'mv';
    $0813	: Result := 'be';
    $0413	: Result := 'nl';
    $0c09	: Result := 'au';
    $2809	: Result := 'be';
    $1009	: Result := 'ca';
    $2409	: Result := '029';
    $4009	: Result := 'in';
    $1809	: Result := 'ie';
    $2009	: Result := 'jm';
    $4409	: Result := 'my';
    $1409	: Result := 'nz';
    $3409	: Result := 'ph';
    $4809	: Result := 'sg';
    $1c09	: Result := 'za';
    $2c09	: Result := 'tt';
    $0809	: Result := 'gb';
    $0409	: Result := 'us';
    $3009	: Result := 'zw';
    $0425	: Result := 'ee';
    $0438	: Result := 'fo';
    $0464	: Result := 'ph';
    $040b	: Result := 'fi';
    $080c	: Result := 'be';
    $0c0c	: Result := 'ca';
    $040c	: Result := 'fr';
    $140c	: Result := 'lu';
    $180c	: Result := 'mc';
    $100c	: Result := 'ch';
    $0462	: Result := 'nl';
    $0456	: Result := 'es';
    $0437	: Result := 'ge';
    $0c07	: Result := 'at';
    $0407	: Result := 'de';
    $1407	: Result := 'li';
    $1007	: Result := 'lu';
    $0807	: Result := 'ch';
    $0408	: Result := 'gr';
    $046f	: Result := 'gl';
    $0447	: Result := 'in';
    $0468	: Result := 'ng';
    $040d	: Result := 'il';
    $0439	: Result := 'in';
    $040e	: Result := 'hu';
    $040f	: Result := 'is';
    $0470	: Result := 'ng';
    $0421	: Result := 'id';
    $085d,$045d
          : Result := 'ca';
    $083c	: Result := 'ie';
    $0410	: Result := 'it';
    $0810	: Result := 'ch';
    $0411	: Result := 'jp';
    $044b	: Result := 'in';
    $043f	: Result := 'kz';
    $0453	: Result := 'kh';
    $0486	: Result := 'gt';
    $0487	: Result := 'rw';
    $0457	: Result := 'in';
    $0412	: Result := 'kr';
    $0440	: Result := 'kg';
    $0454	: Result := 'la';
    $0426	: Result := 'lv';
    $0427	: Result := 'lt';
    $082e	: Result := 'de';
    $046e	: Result := 'lu';
    $042f	: Result := 'mk';
    $083e	: Result := 'bn';
    $043e	: Result := 'my';
    $044c	: Result := 'in';
    $043a	: Result := 'mt';
    $0481	: Result := 'nz';
    $047a	: Result := 'cl';
    $044e	: Result := 'in';
    $047c	: Result := 'ca';
    $0450	: Result := 'mn';
    $0850	: Result := 'cn';
    $0461	: Result := 'np';
    $0414, $0814
          : Result := 'no';
    $0482	: Result := 'fr';
    $0448	: Result := 'in';
    $0463	: Result := 'af';
    $0429	: Result := 'ir';
    $0415	: Result := 'pl';
    $0416	: Result := 'br';
    $0816	: Result := 'pt';
    $0446	: Result := 'in';
    $046b	: Result := 'bo';
    $086b	: Result := 'ec';
    $0c6b	: Result := 'pe';
    $0418	: Result := 'ro';
    $0417	: Result := 'ch';
    $0419	: Result := 'ru';
    $243b	: Result := 'fi';
    $103b	: Result := 'no';
    $143b	: Result := 'se';
    $0c3b	: Result := 'fi';
    $043b	: Result := 'no';
    $083b	: Result := 'se';
    $203b	: Result := 'fi';
    $183b	: Result := 'no';
    $1c3b	: Result := 'se';
    $044f	: Result := 'in';
    $181a	: Result := 'ba';
    $0c1a, $081a
          : Result := 'cs';
    $046c	: Result := 'za';
    $0432	: Result := 'za';
    $045b	: Result := 'lk';
    $041b	: Result := 'sk';
    $0424	: Result := 'si';
    $2c0a	: Result := 'ar';
    $400a	: Result := 'bo';
    $340a	: Result := 'cl';
    $240a	: Result := 'co';
    $140a	: Result := 'cr';
    $1c0a	: Result := 'do';
    $300a	: Result := 'ec';
    $440a	: Result := 'sv';
    $100a	: Result := 'gt';
    $480a	: Result := 'hn';
    $080a	: Result := 'mx';
    $4c0a	: Result := 'ni';
    $180a	: Result := 'pa';
    $3c0a	: Result := 'py';
    $280a	: Result := 'pe';
    $500a	: Result := 'pr';
    $0c0a	: Result := 'es';
    $380a	: Result := 'uy';
    $200a	: Result := 've';
    $0441	: Result := 'ke';
    $081d	: Result := 'fi';
    $041d	: Result := 'se';
    $045a	: Result := 'sy';
    $0428	: Result := 'tj';
    $085f	: Result := 'dz';
    $0449	: Result := 'in';
    $0444	: Result := 'ru';
    $044a	: Result := 'in';
    $041e	: Result := 'th';
    $0451	: Result := 'cn';
    $041f	: Result := 'tr';
    $0442	: Result := 'tm';
    $0480	: Result := 'cn';
    $0422	: Result := 'ua';
    $042e	: Result := 'de';
    $0420	: Result := 'pk';
    $0843	: Result := 'uz';
    $0443	: Result := 'uz';
    $042a	: Result := 'vn';
    $0452	: Result := 'gb';
    $0488	: Result := 'sn';
    $0434	: Result := 'za';
    $0485	: Result := 'ru';
    $0478	: Result := 'cn';
    $046a	: Result := 'ng';
    $0435	: Result := 'za';
  else
    Result:='';
  end;
end;

function GetLangCountry(aLangID:LANGID):string;
begin
  Result:=UpperCase(GetLangSubCode(aLangID));
  if Length(Result)>0 then
    Result:=GetLangPrimaryCode(aLangID)+'-'+Result
  else
    Result:=GetLangPrimaryCode(aLangID);
end;

{$IFNDEF D6}
function DirectoryExists(const Directory: string): Boolean;
var
  Code: Integer;
begin
  Code := GetFileAttributes(PChar(Directory));
  Result := (Code <> -1) and (FILE_ATTRIBUTE_DIRECTORY and Code <> 0);
end;
{$ENDIF}

procedure _GetLangFiles(Dir:string;const PdkFiles,UsrFiles:TStrings;LangCode:string='';Recursive:Boolean=False);
var
	Found:integer;
	SR:TSearchRec;
begin
  {$WARNINGS OFF}
  {$IFDEF D7}
	Dir:=IncludeTrailingPathDelimiter(Dir);
  {$ELSE}
	Dir:=IncludeTrailingBackSlash(Dir);
  {$ENDIF}
  if Recursive then
  begin
    Found := FindFirst(Dir+'*.*', faAnyFile-faVolumeID, SR);
    while Found = 0 do
      begin
        if (sr.Attr and faDirectory>0) and (sr.Name<>'.') and (sr.Name<>'..') then
          _GetLangFiles(dir+sr.Name,PdkFiles,UsrFiles,LangCode,Recursive);
        Found := FindNext(SR);
      end;
    SysUtils.FindClose(SR);
  end;
  if LangCode='' then
    LangCode:='*';

  Found := FindFirst(Dir+'*.'+LangCode+'.'+LANGFILEEXT, faAnyFile-faDirectory-faVolumeID, SR);
  while Found = 0 do
    begin
      PdkFiles.Add(Dir+sr.Name);
      Found := FindNext(SR);
    end;
  SysUtils.FindClose(SR);

  Found := FindFirst(Dir+'*.'+LangCode+'.'+LANGFILEEXTZ, faAnyFile-faDirectory-faVolumeID, SR);
  while Found = 0 do
    begin
      PdkFiles.Add(Dir+sr.Name);
      Found := FindNext(SR);
    end;
  SysUtils.FindClose(SR);

  Found := FindFirst(Dir+'*.'+LangCode+'.'+LANGFILEEXTU, faAnyFile-faDirectory-faVolumeID, SR);
  while Found = 0 do
    begin
      UsrFiles.Add(Dir+sr.Name);
      Found := FindNext(SR);
    end;
  SysUtils.FindClose(SR);

  {$WARNINGS ON}
end;

procedure _EscapeText(var Txt:string);
begin
  if AnsiPos(V_CRLF,Txt)>0 then
    Txt:=StringReplace(Txt,V_CRLF,CRLF,[rfReplaceAll]);
  if AnsiPos(V_TAB,Txt)>0 then
    Txt:=StringReplace(Txt,V_TAB,TAB,[rfReplaceAll]);
end;

function _ECodeText(const Txt:string):string;
begin
  Result:=Txt;
  if AnsiPos(CRLF,Result)>0 then
    Result:=StringReplace(Result,CRLF,V_CRLF,[rfReplaceAll]);
  if AnsiPos(Tab,Result)>0 then
    Result:=StringReplace(Result,TAB,V_TAB,[rfReplaceAll]);
end;

{$IFNDEF D12}
function LoadUTF8File(var AString: {$IFDEF D7}UTF8String{$ELSE}AnsiString{$ENDIF}; const FileName: string; FileMode: word = fmOpenRead or fmShareDenyNone): boolean;
var
  L: integer;
  FS: TFileStream;
  preamb:string;
begin
  Result:=False;
  try
    FS := TFileStream.Create(FileName, FileMode);
    try
      L := FS.Size-3;
      if L > 0 then
      begin
        SetLength(preamb,3);
        FS.ReadBuffer(Pointer(preamb)^, 3);
        if preamb=#$EF#$BB#$BF then
        begin
          SetLength(AString, L);
          FS.ReadBuffer(Pointer(AString)^, L);
          Result := true;
        end;
      end;
    finally
      FS.free
    end;
  except
    Result := false;
  end;
end;
{$ENDIF}

{-----------------------------------------------------------------------------
  TLanguageManager
-----------------------------------------------------------------------------}

constructor TLanguageManager.Create;
begin
	inherited;
  fLangProps:=TStringList.Create;
  fLangConsts:=TStringList.Create;
  fEMessList:=TStringList.Create;
  fEMessConv:=TStringList.Create;
  fLangResStr:=TStringList.Create;
  fFileNames:=TStringList.Create;
  fActiveFiles:=TStringList.Create;
  fFolder:=ExtractFilePath(ParamStr(0))+LANGDIR_RELPATH;
  fOnLanguageChanged:=nil;
  fAllowSpecialProps:=False;
  fRecursiveFolder:=False;
  fLanguageCode:='';
  fLanguageName:='';
  fBiDiMode:=bdLeftToRight;
  fLastError:='';
end;

destructor TLanguageManager.Destroy;
begin
  Clear;
  fActiveFiles.Free;
  fFileNames.Free;
  fLangResStr.Free;
  fEMessList.Free;
  fEMessConv.Free;
  fLangProps.Free;
  fLangConsts.Free;
  inherited;
end;

procedure TLanguageManager.Clear(Consts:Boolean=true;Props:Boolean=True;Messages:Boolean=False);
var
  i:Integer;
  n:^String;
begin
  if Consts then
  begin
    for i:=0 to fLangConsts.Count-1 do
    begin
      n:=Pointer(fLangConsts.Objects[i]);
      Dispose(n);
    end;
    fLangConsts.Clear;
  end;
  if Props then
  begin
    for i:=0 to fLangProps.Count-1 do
    begin
      n:=Pointer(fLangProps.Objects[i]);
      Dispose(n);
    end;
    fLangProps.Clear;
  end;
  if Messages then
  begin
    fEMessList.Clear;
    fEMessConv.Clear;
  end;
  for i:=0 to fLangResStr.Count-1 do
  begin
    n:=Pointer(fLangResStr.Objects[i]);
    Dispose(n);
  end;
  fLangResStr.Clear;
  fActiveFiles.Clear;
end;

procedure TLanguageManager.FilterFileNames(const FileList:TStrings);
var
  i,j:Integer;
  f:Boolean;
begin
  if fFileNames.Count>0 then
    for i := FileList.Count - 1 downto 0 do
    begin
      f:=False;
      for j := 0 to fFileNames.Count - 1 do
        if MatchesMask(ExtractFileName(FileList[i]),fFileNames[j]+'.*.*') then
        begin
          f:=True;
          Break;
        end;
      if not f then
        FileList.Delete(i);
    end;
end;

function TLanguageManager.GetAvailableLanguages(const AvailableLanguages:TStringList):Boolean;
var
  i:Integer;
  PF,UF:TStringList;
  Ini:TMemIniFile;
  lc,lng:string;
  {$IFNDEF D12}
    {$IFDEF D7}
  u:UTF8String;
  WS:TStringList;
    {$ENDIF}
  {$ENDIF}
  {$IFDEF PLS_LNGZ}
  Zip:TZipFile;
  {$ENDIF}
begin
  Result:=False;
  fLastError:='';
  try
    PF:=TStringList.Create;
    UF:=TStringList.Create;
    try
      AvailableLanguages.Clear;
      // get all langauge filenames in language folder
      if (fFolder<>'') and DirectoryExists(fFolder) then
        _GetLangFiles(fFolder,PF,UF,'',fRecursiveFolder)
      else
      begin
        _GetLangFiles(ExtractFilePath(ParamStr(0))+LANGDIR_RELPATH,PF,UF,'',fRecursiveFolder);
        if PF.Count+UF.Count=0 then
          _GetLangFiles(ExtractFilePath(ParamStr(0)),PF,UF,'',fRecursiveFolder);
      end;
      // filter by user defined filename identifiers
      PF.AddStrings(UF); // here it is not important to differentiate
      FilterFileNames(PF);
      for i := 0 to PF.Count - 1 do
      begin
        // extract language code from filename
        lc:=AnsiLowerCase(ExtractFileExt(ChangeFileExt(PF[i],'')));
        Delete(lc,1,1);
        if AvailableLanguages.IndexOfName(lc)=-1 then
        begin
          lng:='';
          try
            if lowercase(ExtractFileExt(PF[i]))='.'+LANGFILEEXTZ then
            begin
              {$IFDEF PLS_LNGZ}
              Zip:=TZipFile.Create;
              try
                Zip.LoadFromFile(PF[i]);
                {$IFDEF D7}
                  lng:=UTF8Decode(Zip.ZipFileComment);
                {$ELSE}
                  lng:=Zip.ZipFileComment;
                {$ENDIF}
              finally
                Zip.Free;
              end;
              {$ENDIF}
            end
            else
            begin
              {$IFDEF D12}
              Ini:=TMemIniFile.Create(PF[i]);
              {$ELSE}
                {$IFDEF D7}
              if LoadUTF8File(u,PF[i]) then
              begin
                Ini:=TMemIniFile.Create('');
                WS:=TStringList.Create;
                try
                  WS.Text:=UTF8Decode(u);
                  Ini.SetStrings(WS);
                finally
                  WS.Free;
                end;
              end
              else
                Ini:=TMemIniFile.Create(PF[i]);
                {$ELSE}
              Ini:=TMemIniFile.Create(PF[i]);
                {$ENDIF}
              {$ENDIF}
              try
                lng:=Ini.ReadString(lini_sec_main,lini_key_name,'');
              finally
                Ini.Free;
              end;
            end;
          except
          end;
          if lng<>'' then
            AvailableLanguages.Add(lc+'='+lng);
        end;
      end;
      Result:=True;
    finally
      UF.Free;
      PF.Free;
    end;
  except
    on E:Exception do
      fLastError:=e.Message;
  end;
end;

procedure TLanguageManager.GetLanguageFiles(LngCode:string; const PdkFiles, UsrFiles:TStrings);
begin
  if (fFolder<>'') and DirectoryExists(fFolder) then
    _GetLangFiles(fFolder,PdkFiles,UsrFiles,LngCode,fRecursiveFolder)
  else
  begin
    _GetLangFiles(ExtractFilePath(ParamStr(0))+LANGDIR_RELPATH,PdkFiles,UsrFiles,LngCode,fRecursiveFolder);
    if PdkFiles.Count+UsrFiles.Count=0 then
      _GetLangFiles(ExtractFilePath(ParamStr(0)),PdkFiles,UsrFiles,LngCode,fRecursiveFolder);
  end;
  FilterFileNames(PdkFiles); FilterFileNames(UsrFiles); // filter by user defined filename identifiers
end;

function TLanguageManager.ReadLanguageData(Ini:TMemIniFile; IsUser:Boolean; var LangName:string):Boolean;
var
  sn,fn:String;
	n:^String;
  i,j,k:Integer;
  SL,SLN:TStringList;
begin
  Result:=False;
  if Assigned(Ini) then
  begin
    SL:=TStringList.Create;
    SLN:=TStringList.Create;
    try
      if LangName='' then
        LangName:=Ini.ReadString(lini_sec_main,lini_key_name,'');
      if (fBiDiMode<>bdRightToLeft) and Ini.ReadBool(lini_sec_main,lini_key_rtl,False) then
        fBiDiMode:=bdRightToLeft;

      // load constants
      Ini.ReadSectionValues(lini_sec_consts,SL);
      for i:=0 to SL.Count-1 do
      begin
        sn:=AnsiUpperCase(SL.Names[i]);
        if Length(SN)>0 then
        begin
          if SN[1]='@' then // it is a resourcestring constant
          begin
            if isUser and fLangResStr.Find(sn,k) then
            begin
              n:=Pointer(fLangResStr.Objects[k]);
              {$IFDEF D7}
              n^:=SL.ValueFromIndex[i];
              {$ELSE}
              n^:=SL.Values[sn];
              {$ENDIF}
              _EscapeText(n^);
            end
            else
            begin
              new(n);
              {$IFDEF D7}
              n^:=SL.ValueFromIndex[i];
              {$ELSE}
              n^:=SL.Values[sn];
              {$ENDIF}
              _EscapeText(n^);
              fLangResStr.AddObject(sn,Pointer(n));
            end;
          end
          else
          begin
            if isUser and fLangConsts.Find(sn,k) then
            begin
              n:=Pointer(fLangConsts.Objects[k]);
              {$IFDEF D7}
              n^:=SL.ValueFromIndex[i];
              {$ELSE}
              n^:=SL.Values[sn];
              {$ENDIF}
              _EscapeText(n^);
            end
            else
            begin
              new(n);
              {$IFDEF D7}
              n^:=SL.ValueFromIndex[i];
              {$ELSE}
              n^:=SL.Values[sn];
              {$ENDIF}
              _EscapeText(n^);
              fLangConsts.AddObject(sn,Pointer(n));
            end;
          end;
        end;
      end;

      // load properties
      Ini.ReadSections(SLN);
      SLN.Sorted:=True;
      {$IFDEF D7}
      SLN.CaseSensitive:=False;
      {$ENDIF}
      i:=SLN.IndexOf(lini_sec_main); if i>=0 then SLN.Delete(i);
      i:=SLN.IndexOf(lini_sec_consts); if i>=0 then SLN.Delete(i);
      i:=SLN.IndexOf(lini_sec_messages); if i>=0 then SLN.Delete(i);
      for j := 0 to SLN.Count - 1 do
      begin
        SL.Clear;
        Ini.ReadSectionValues(SLN[j],SL);
        fn:=AnsiUpperCase(SLN[j])+'.';
        for i:=0 to SL.Count-1 do
        begin
          sn:=AnsiUpperCase(SL.Names[i]);
          if isUser and fLangProps.Find(fn+sn,k) then
          begin
            n:=Pointer(fLangProps.Objects[k]);
           {$IFDEF D7}
            n^:=SL.ValueFromIndex[i];
            {$ELSE}
            n^:=SL.Values[sn];
            {$ENDIF}
            _EscapeText(n^);
          end
          else
          begin
            new(n);
            {$IFDEF D7}
            n^:=SL.ValueFromIndex[i];
            {$ELSE}
            n^:=SL.Values[sn];
            {$ENDIF}
            _EscapeText(n^);
            fLangProps.AddObject(fn+sn,Pointer(n));
          end;
        end;
      end;

      // load error messages
      SL.Clear;
      Ini.ReadSectionValues(lini_sec_messages,SL);
      i:=0;
      while i<SL.Count do
      begin
        fn:=trim(SL[i]);
        if (Length(fn)>0) and (Copy(fn,1,1)<>';') then
        begin
          Inc(i);
          if i<SL.Count then
          begin
            _EscapeText(fn);
            sn:=SL[i];
            _EscapeText(sn);
            if isUser then
              j:=fEMessList.IndexOf(fn)
            else
              j:=-1;
            if j>=0 then
              fEMessConv[j]:=sn
            else
            begin
              fEMessList.Add(fn);
              fEMessConv.Add(sn);
            end;
          end;
        end;
        Inc(i);
      end;
      Result:=True;
    finally
      SLN.Free;
      SL.Free;
    end;
  end;
end;

function TLanguageManager.LoadLanguageFile(const FileName:string;var LangName:string;var ErrMsg:string):Boolean;
var
  Ini:TMemIniFile;
  {$IFNDEF D12}
    {$IFDEF D7}
  u:UTF8String;
  WS:TStringList;
    {$ENDIF}
  {$ENDIF}
  {$IFDEF PLS_LNGZ}
  Zip:TZipFile;
  i:Integer;
  MS:TMemoryStream;
  ZS:TStringList;
  d:AnsiString;
  {$ENDIF}
begin
  Result:=False;
  ErrMsg:='';
  try
    if lowercase(ExtractFileExt(FileName))='.'+LANGFILEEXTZ then
    begin
      {$IFDEF PLS_LNGZ}
      Zip:=TZipFile.Create;
      MS:=TMemoryStream.Create;
      ZS:=TStringList.Create;
      try
        Zip.LoadFromFile(FileName);
        Result:=True;
        for i:=0 To Zip.Count-1 do
        begin
          Ini:=TMemIniFile.Create('');
          d:=Zip.Data[i];
          MS.Clear;
          ZS.Clear;
          try
            {$IFDEF D12}
            MS.Write(d[1],Length(d));
            MS.Position:=0;
            ZS.LoadFromStream(MS);
            {$ELSE}
              {$IFDEF D7}
              if Copy(d,1,3)=#$EF#$BB#$BF then
                Delete(d,1,3);
              ZS.Text:=UTF8Decode(d);
              {$ELSE}
              ZS.Text:=d;
              {$ENDIF}
            {$ENDIF}
            Ini.SetStrings(ZS);
            if not ReadLanguageData(Ini,AnsiLowerCase(ExtractFileExt(Zip.Name[i]))='.'+LANGFILEEXTU,LangName) then
              Result:=False;
          finally
            Ini.Free;
          end;
        end;
      finally
        ZS.Free;
        MS.Free;
        Zip.Free;
      end;
      {$ENDIF}
    end
    else
    begin
      {$IFDEF D12}
      Ini:=TMemIniFile.Create(FileName);
      {$ELSE}
        {$IFDEF D7}
      if LoadUTF8File(u,FileName) then
      begin
        Ini:=TMemIniFile.Create('');
        WS:=TStringList.Create;
        try
          WS.Text:=UTF8Decode(u);
          Ini.SetStrings(WS);
        finally
          WS.Free;
        end;
      end
      else
        Ini:=TMemIniFile.Create(FileName);
        {$ELSE}
      Ini:=TMemIniFile.Create(FileName);
        {$ENDIF}
      {$ENDIF}
      try
        Result:=ReadLanguageData(Ini,AnsiLowerCase(ExtractFileExt(FileName))='.'+LANGFILEEXTU,LangName);
      finally
        Ini.Free;
      end;
    end;
    fActiveFiles.Add(FileName);
  except
    on E:Exception do
      ErrMsg:=e.Message;
  end;
end;

procedure TLanguageManager.SetLanguageCode(const Value:string);
var
  i:Integer;
  UF,PF:TStringList;
  err:string;
begin
  fLastError:='';
  if Value<>'' then
  begin
    try
      Clear(True,True,True);
      fLanguageCode:=Value;
      fLanguageName:='';
      fBiDiMode:=bdLeftToRight;
      fLangConsts.Sorted:=False;
      fLangProps.Sorted:=False;
      fLangResStr.Sorted:=False;
      PF:=TStringList.Create;
      UF:=TStringList.Create;
      try
        // get all product (PF) and user (UF) langauge filenames in language folder
        if (fFolder<>'') and DirectoryExists(fFolder) then
          _GetLangFiles(fFolder,PF,UF,Value,fRecursiveFolder)
        else
        begin
          _GetLangFiles(ExtractFilePath(ParamStr(0))+LANGDIR_RELPATH,PF,UF,Value,fRecursiveFolder);
          if PF.Count+UF.Count=0 then
            _GetLangFiles(ExtractFilePath(ParamStr(0)),PF,UF,Value,fRecursiveFolder);
        end;
        FilterFileNames(PF); FilterFileNames(UF); // filter by user defined filename identifiers

        if PF.count+UF.Count=0 then
        begin
          fLanguageCode:='';
          raise Exception.Create(resELangFilesNotFound);
        end;

        if PF.Count>0 then
        begin
          for i := 0 to PF.Count - 1 do
          begin
            if not LoadLanguageFile(PF[i],fLanguageName,err) then
              fLastError:=fLastError+CRLF+err
          end;
        end;
        if UF.Count>0 then
        begin
          fLangConsts.Sorted:=True;
          fLangProps.Sorted:=True;
          fLangResStr.Sorted:=True;
          for i := 0 to UF.Count - 1 do
          begin
            if not LoadLanguageFile(UF[i],fLanguageName,err) then
              fLastError:=fLastError+CRLF+err
          end;
        end;
        Delete(fLastError,1,2);
      finally
        if UF.Count=0 then
        begin
          fLangConsts.Sorted:=True;
          fLangProps.Sorted:=True;
          fLangResStr.Sorted:=True;
        end;
        UF.Free;
        PF.Free;
      end;
    except
      on E:Exception do
        fLastError:=fLastError+e.Message;
    end;
    if Assigned(fOnLanguageChanged) and (fLastError='') and (fLanguageCode=Value) then
      fOnLanguageChanged(Self);
  end;
end;

procedure TLanguageManager.Refresh;
begin
  LanguageCode:=fLanguageCode;
end;

function TLanguageManager.LangText(const ID:string;const DefaultText:string=''):string;
var
	i:Integer;
begin
  if Length(ID)=0 then
  begin
    Result:=DefaultText;
    _EscapeText(Result);
  end
  else
  begin
    if ID[1]='@' then // it is a resourcestring constant
    begin
      if fLangResStr.Find(UpperCase(ID),i) then
        Result:=PChar(Pointer(fLangResStr.Objects[i])^)
      else
      begin
        Result:=DefaultText;
        _EscapeText(Result);
      end
    end
    else
    if fLangConsts.Find(UpperCase(ID),i) then
      Result:=PChar(Pointer(fLangConsts.Objects[i])^)
    else
    begin
      if fLangProps.Find(UpperCase(ID),i) then
        Result:=PChar(Pointer(fLangProps.Objects[i])^)
      else
      begin
        Result:=DefaultText;
        _EscapeText(Result);
      end;
    end;
  end;
end;

function TLanguageManager.LangVCL(const Owner:TComponent;const AsClassName:String='';const OnlyComponent:TComponent=nil):Boolean;
var
	i,j,p1,p2,c2,c3:Integer;
	fm,pn,cln,cls:String;
	C:TComponent;
	CL:TCollectionItem;
  Coll:TCollection;
	GO:TObject;
	pvV:Variant;
  pvI:Integer;
	pv:String;
begin
  fLastError:='';
	try
		if AsClassName<>'' then
			fm:=AnsiUpperCase(AsClassName)+'.' // form/frame classname
		else
			fm:=AnsiUpperCase(Owner.ClassName)+'.';
		p1:=Length(fm);
		for i:=0 to fLangProps.Count-1 do
			if Copy(fLangProps[i],1,p1)=fm then
			begin // form found
				for j:=i to fLangProps.Count-1 do
				begin
					if Copy(fLangProps[j],1,p1)<>fm then
            break; // break on next form
          try
            p2:=LastDelimiter('.',fLangProps[j]);
            cln:=Copy(fLangProps[j],p1+1,p2-p1-1);
            c2:=Pos('.',cln);
            if c2>0 then
            begin // child object of component (i.e. Collection...)
              cls:=Copy(cln,1,c2-1);
              if CompareText(cls,LC_Self)=0 then
                c:=Owner
              else
                c:=Owner.findcomponent(cls); // component
              Delete(cln,1,c2); // now cln = child object name
              c2:=Pos('.',cln);
              if c2>0 then
              begin
                repeat
                  cls:=Copy(cln,1,c2-1);
                  GO:=TObject(GetOrdProp(c,cls));
                  Delete(cln,1,c2);
                  c2:=Pos('.',cln);
                until c2=0;
                c:=TComponent(GO);
              end;
              if (OnlyComponent<>nil) and (c<>OnlyComponent) then
                continue;
							if (c<>nil) then
              begin
                c2:=Pos('[',cln);
                if c2>0 then
                begin // item of object specified (i.e. collectionitem)
                  c3:=strtointdef(Copy(cln,c2+1,Pos(']',cln)-c2-1),-1);
                  if c3>-1 then
                  begin
                    Coll:=TCollection(GetOrdProp(c,Copy(cln,1,c2-1)));
                    CL:=Coll.FindItemID(c3);
                    if (CL=nil) and (c3<Coll.Count) then
                      CL:=Coll.Items[c3];
                    //if Assigned(CL) then
                    begin
                      pn:=Copy(fLangProps[j],p2+1,maxint); // property name
                      pv:=PChar(pointer(fLangProps.Objects[j])^); // property value
                      SetPropValue(CL,pn,pv);
                    end;
                  end;
                end
                else
                begin // whole subobject
                  GO:=TObject(GetOrdProp(c,cln));
                  pn:=Copy(fLangProps[j],p2+1,maxint); // property name
                  // solve not published properties, that are commonly used
                  if GO is TStrings then
                  begin
                    if pn='TEXT' then
                      TStrings(GO).Text:=PChar(Pointer(fLangProps.Objects[j])^)
                    else
                    if pn='COMMATEXT' then
                      TStrings(GO).CommaText:=PChar(Pointer(fLangProps.Objects[j])^)
                    {$IFDEF D7}
                    else
                    if pn='DELIMITEDTEXT' then
                      TStrings(GO).DelimitedText:=PChar(Pointer(fLangProps.Objects[j])^)
                    {$ENDIF}
                  end
                  else
                  begin // for other objects
                    pvV:=String(PChar(pointer(fLangProps.Objects[j])^)); // property value
                    SetPropValue(GO,pn,pvV);
                  end;
                end;
              end;
            end
            else
            begin // standard component
              if CompareText(cln,LC_Self)=0 then
                c:=Owner
              else
                c:=Owner.findcomponent(cln); // component
              if (OnlyComponent<>nil) and (c<>OnlyComponent) then
                continue;
              if c<>nil then
              begin
                pn:=Copy(fLangProps[j],p2+1,maxint); // property name
                pv:=PChar(pointer(fLangProps.Objects[j])^); // property value
                // some special handled properties
                if FAllowSpecialProps then
                begin
                  if (pn='SHORTCUT') then
                  begin
                    pvI:=TextToShortCut(pv);
                    SetOrdProp(c,pn,pvI);
                  end
                  else
                    SetPropValue(c,pn,pv);
                end
                else
                  SetPropValue(c,pn,pv);
							end
            end;
					except
            on E:Exception do
              fLastError:=fLastError+CRLF+CRLF+Self.ClassName+' ('+fLangProps[j]+'):'+CRLF+'      '+e.Message;
          end;
        end;
				break;
			end;
  except
    on E2:Exception do
      fLastError:=fLastError+CRLF+CRLF+Self.ClassName+':'+CRLF+'      '+e2.Message;
  end;
  Result:=fLastError='';
//  Delete(fLastError,1,2);
end;

function _StrLastPos(const SubStr, S: string): Integer;
var
  Last, Current: PChar;
begin
  Result := 0;
  Last := nil;
  Current := PChar(S);

  while (Current <> nil) and (Current^ <> #0) do
  begin
    Current := StrPos(PChar(Current), PChar(SubStr));
    if Current <> nil then
    begin
      Last := Current;
      Inc(Current);
    end;
  end;
  if Last <> nil then
    Result := Abs(PChar(S) - Last) + 1;
end;

function TLanguageManager.LangMess(const aMessage:string):string; // Returns localized message converted via mask or returns itself
var
	i:Integer;
  pl,pc,pu:Integer;
  eml,emc,emu:string;
  f:Boolean;
begin
  f:=False;
  // identify message by mask
  for i:=0 To fEMessList.Count-1 do
  begin
    eml:=fEMessList[i];
    if MatchesMask(aMessage,eml) then
    begin
      // convert message via translated mask
      emc:=fEMessConv[i];
      emu:=aMessage;
      Result:='';
      pc:=LastDelimiter('*?',emc);
      while pc>0 do
      begin
        Result:=Copy(emc,pc+1,maxint)+Result;
        Delete(emc,pc,maxint);
        pl:=LastDelimiter('*?',eml);
        Delete(emu,Length(emu)-Length(Copy(eml,pl+1,maxint))+1,maxint);
        Delete(eml,pl,maxint);

        pc:=LastDelimiter('*?',emc);
        if pc=0 then
        begin
          Delete(emu,1,Length(eml));
          Result:=emu+Result;
        end
        else
        begin
          pl:=LastDelimiter('*?',eml);
          pu:=_StrLastPos(Copy(eml,pl+1,maxint),emu);
          Result:=Copy(emu,pu+Length(eml)-pl,maxint)+Result;
          Delete(emu,pu+Length(eml)-pl,MaxInt);
        end;
      end;
      if Length(emc)>0 then
        Result:=emc+Result;
      f:=True;
      Break;
    end;
  end;
  if not f then
    Result:=aMessage;
end;

// LangResourceStr - author of the algorithm for this method is TeamB (www.teamb.com)
procedure TLanguageManager.LangResourceStr(rs: PResStringRec; newStr: PChar);
var
  oldprotect: DWORD;
begin
  If rs = nil Then Exit;
  Assert( Assigned(newStr), 'newStr must not be nil');
  VirtualProtect(rs, SizeOf(rs^), PAGE_EXECUTE_READWRITE, @oldProtect);
  rs^.Identifier := Integer(newStr);
  VirtualProtect(rs, SizeOf(rs^), oldProtect, @oldProtect);
end;

initialization

  if not IsLibrary then
		LanguageManager:=TLanguageManager.Create;

finalization

  if (not IsLibrary) and Assigned(LanguageManager) then
    LanguageManager.Free;

end.


